
import os
import sys
import socket
import subprocess
import time
import re
import string
import io
import tempfile

sys.path.append(os.environ['ANTELOPE'] + "/data/python")

from PIL import Image  
from PIL import XwdImagePlugin  
import antelope.orb as orb  
import antelope.stock as stock  
import antelope.elog as elog  
import _xwrap  


class RtcacheEngine(object):
    """ Object to manage one stream of cached entities within the rtcache program"
    """

    streams = dict()
    xwindow_displays = dict()
    next_displaynumber_guess = 1
    tabstop = 8

    @staticmethod
    def _reindent(s, numSpaces):

        s = string.split(s, '\n')

        offset = -1

        for line in s:

            line = re.sub('\t', ' ' * RtcacheEngine.tabstop, line)

            if len(line) > 0 and not re.match('^\s*$', line):
                i = 0
                while i < len(line) and line[i] == ' ':
                    i = i + 1
                if offset == -1 or i < offset:
                    offset = i

        s = [_l.replace('\t', ' ' * RtcacheEngine.tabstop) for _l in s]
        s = [_l.replace(offset * ' ', numSpaces * ' ', 1) for _l in s]

        s = string.join(s, '\n')

        return s

    def inform(self, message):

        if self.verbose:
            elog.notify("RtcacheEngine '%s': %s" % (self.streamname, message))

    def _diagnose(self, message):

        if self.debug:
            elog.notify("RtcacheEngine diagnostic: '%s': %s" % (self.streamname, message))

    def _displayport_available(self, display):

        xoffset = 6000

        if type(display) == str:

            if display[0] == ':':

                display = display[1:]

            display = int(display)

        iport = xoffset + display

        lockfile = "/tmp/.X%d-lock" % display

        if os.path.exists(lockfile):

            self._diagnose("_displayport_available: test results for port %d: unavailable, lockfile '%s' exists" %
                           (iport, lockfile))
            return False

        socketfile = "/tmp/.X11-unix/X%d" % display

        if os.path.exists(socketfile):

            self._diagnose("_displayport_available: test results for port %d: unavailable, socketfile '%s' exists" %
                           (iport, socketfile))
            return False

        try:

            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            s.connect(('localhost', iport))

            self._diagnose("_displayport_available: test results for port %d: Port %d is unavailable" % (iport, iport))

            return False

        except socket.error:

            self._diagnose("_displayport_available: test results for port %d: Port %d is available" % (iport, iport))

            return True

    def _next_open_display(self):

        max_attempts = 100
        max_displaynumber_guess = 1000

        while max_attempts > 0 and RtcacheEngine.next_displaynumber_guess < max_displaynumber_guess:

            if self._displayport_available(RtcacheEngine.next_displaynumber_guess):

                opendisplay = RtcacheEngine.next_displaynumber_guess

                RtcacheEngine.next_displaynumber_guess = RtcacheEngine.next_displaynumber_guess + 1

                return ":%d" % opendisplay

            else:

                if RtcacheEngine.next_displaynumber_guess >= max_displaynumber_guess:

                    RtcacheEngine.next_displaynumber_guess = 1

                else:

                    RtcacheEngine.next_displaynumber_guess = RtcacheEngine.next_displaynumber_guess + 1

        elog.complain("RtcacheEngine '%s': Failed to automatically find display port after %d attempts, giving up" %
                      (self.streamname, max_attempts))

        return None

    def _init_engine_diy(self):

        expected = ['startup_script', 'update_script', 'harvest_script', 'shutdown_script']

        for param in expected:

            if param not in self.pfdict_config:

                raise RuntimeError("RtcacheEngine: Failed to start diy engine for stream name '%s': "
                                   "missing parameter '%s'" % (self.streamname, param))
            else:
                cmd = "self.%s = self.pfdict_config['%s']" % (param, param)

                exec(cmd)

        self.startup_script = RtcacheEngine._reindent(self.startup_script, 0)
        self.update_script = RtcacheEngine._reindent(self.update_script, 0)
        self.harvest_script = RtcacheEngine._reindent(self.harvest_script, 0)
        self.shutdown_script = RtcacheEngine._reindent(self.shutdown_script, 0)

    def _startup_diy(self):

        exec(self.startup_script)

    def _update_diy(self):

        exec(self.update_script)

    def _harvest_diy(self, pfdict_request):

        exec(self.harvest_script)

    def _shutdown_diy(self):

        exec(self.shutdown_script)

    def xdisplay_setup(self, display, screen_number, geometry, startup_sleep_sec):

        if display == "auto":

            self.virtual_display = self._next_open_display()

            if self.virtual_display is None:

                elog.complain("RtcacheEngine: Failed to configure display automatically")

                raise RuntimeError("xdisplay_setup: Failed to initiate X display for streamname '%s'" % self.streamname)

            else:

                self.inform("Automatically configured virtual display '%s' for streamname '%s'" %
                            (self.virtual_display, self.streamname))

        else:

            if not self._displayport_available(self.virtual_display):

                elog.complain("RtcacheEngine: display '%s' is not available" % self.virtual_display)

                raise RuntimeError("xdisplay_setup: Failed to initiate X display for streamname '%s'" % self.streamname)

            else:

                self.virtual_display = display

        if geometry == "":
            self.virtual_screen_geometry = "1280x1024"
        else:
            self.virtual_screen_geometry = geometry

        for e in ['Xvfb', 'xwd']:

            path = stock.datafile('PATH', e)

            if not (type(path) == str and os.path.isfile(path) and os.access(path, os.X_OK)):
                raise RuntimeError("RtcacheEngine: can't find executable '%s'" % e)

        geometry = "%sx24" % self.virtual_screen_geometry

        self.xvfb = subprocess.Popen("Xvfb %s -screen %s %s" % (self.virtual_display, screen_number, geometry),
                                     shell=True)

        time.sleep(float(startup_sleep_sec))

        self.xvfb.poll()

        if self.xvfb.returncode is not None:
            raise RuntimeError("RtcacheEngine: Failed to launch Xvfb: return code %d" % self.xvfb.returncode)
        else:
            self._diagnose("xdisplay_setup: Launched Xvfb as pid %i\n" % self.xvfb.pid)

        if self.virtual_display in RtcacheEngine.xwindow_displays:

            raise RuntimeError("RtcacheEngine: display '%s' already registered to stream name '%s'" %
                               (self.virtual_display, RtcacheEngine.xwindow_displays[display]))
        else:
            RtcacheEngine.xwindow_displays[self.virtual_display] = self.streamname
        return

    def xdisplay_launchcmd(self, command):

        self.command_env['DISPLAY'] = self.virtual_display

        if type(command) == str:
            useshell = True
        else:
            useshell = False

        self._diagnose("xdisplay_launchcmd: About to launch command '%s' under env %r" % (str(command),
                                                                                          self.command_env))

        self.subproc = subprocess.Popen(command, env=self.command_env, stdin=subprocess.PIPE, shell=useshell)

        time.sleep(float(self.startup_sleep_sec))

        self.subproc.poll()

        if self.subproc.returncode is not None:

            raise RuntimeError("RtcacheEngine: Failed to launch command '%r' for stream name '%s': return code %d" %
                               (command, self.streamname, self.subproc.returncode))

        self.inform("startup_xwindow: Launched command '%s' as pid %i\n" % (command, self.subproc.pid))

        time.sleep(float(self.startup_sleep_sec))

    def _find_xwindow_idmatch(self):

        self._diagnose("_find_xwindow_idmatch: Finding application window IDs\n")

        idmatch = list()

        d = _xwrap._XOpenDisplay(self.virtual_display)

        if d is None:
            self._diagnose("_find_xwindow_idmatch: Failed to open display '%s'\n" % self.virtual_display)
            return None

        r = _xwrap._RootWindow(d, 0)

        if r == 0:

            self._diagnose("_find_xwindow_idmatch: Failed to retrieve root window from '%s'\n" % self.virtual_display)

            _xwrap._XCloseDisplay(d)

            return None

        (rc, root, parent, children) = _xwrap._XQueryTree(d, r)

        if rc < 0:
            self._diagnose("_find_xwindow_idmatch: Failed to query root window for children on display '%s'\n" %
                           self.virtual_display)
            _xwrap._XCloseDisplay(d)

            return None

        for child in children:

            (rc, win_attributes) = _xwrap._XGetWindowAttributes(d, child)

            if win_attributes['map_state'] != _xwrap.IsViewable:
                self._diagnose("_find_xwindow_idmatch: Child 0x%x is not viewable, skipping" % child)
                continue

            if self.window_name != '':

                (rc, name) = _xwrap._XGetWMName(d, child)

                if name is None or name == "" or not re.match(self.window_name, name):

                    self._diagnose("_find_xwindow_idmatch: Child 0x%x doesn't match requested name '%s', skipping" %
                                   (child, self.window_name))

                    continue

            (rc, resname, resclass) = _xwrap._XGetClassHint(d, child)

            if rc >= 0 and resname == 'icon':

                self._diagnose("_find_xwindow_idmatch: Child 0x%x is an icon, skipping" % child)
                continue

            if rc >= 0 and resname == 'balloon':

                self._diagnose("_find_xwindow_idmatch: Child 0x%x is an icon, skipping" % child)
                continue

            elif resclass == 'Dialog':

                idmatch.insert(0, child)
                continue

            elif resclass == 'Menu':

                idmatch.insert(0, child)
                continue

            elif resclass == 'Toplevel':

                idmatch.insert(0, child)
                continue

            (rc, root, rel_x, rel_y, width, height, border, depth) = _xwrap._XGetGeometry(d, child)

            if rc < 0:
                self._diagnose("_find_xwindow_idmatch: Child 0x%x doesn't have geometry info), skipping" % child)
                continue

            if width < 8 or height < 8:
                self._diagnose("_find_xwindow_idmatch: Child 0x%x is too small (%dx%d), skipping" %
                               (child, width, height))
                continue

            self._diagnose("_find_xwindow_idmatch: Adding child 0x%x" % child)

            idmatch.append(child)

        _xwrap._XCloseDisplay(d)

        self._diagnose("_find_xwindow_idmatch: Obtained child window matches of %r\n" % idmatch)

        return idmatch

    def xdisplay_harvest(self, pfdict=None):

        self.inform("xdisplay_harvest: starting image harvest at '%s'\n" % stock.epoch2str(stock.now(), "%D %T %Z"))

        if not self.xdisplay_is_running():

            self.inform("xdisplay_harvest: display not running; restarting Xvfb and X client at '%s'\n" %
                        stock.epoch2str(stock.now(), "%D %T %Z"))

            self._update_xwindow(force=True)

        idmatch = None

        remaining = 3

        while (idmatch is None or len(idmatch) <= 0) and remaining > 0:

            idmatch = self._find_xwindow_idmatch()

            if idmatch is None:

                elog.complain("RtcacheEngine: Attempt to find xwindow failed, sleeping %s sec and trying again" %
                              self.startup_sleep_sec)

                time.sleep(float(self.startup_sleep_sec))

                idmatch = self._find_xwindow_idmatch()

            remaining -= 1

        if idmatch is None or len(idmatch) <= 0:

            raise RuntimeError("RtcacheEngine: Failed to find xwindow in virtual display '%s' for stream name '%s'." %
                               (self.virtual_display, self.streamname))

        xwinid = "0x%x" % idmatch[0]

        if pfdict and 'action' in pfdict:

            if pfdict['action'] == 'mouseclick':

                if 'x' in pfdict and re.match('^\d+$', pfdict['x']):

                    rel_click_x = int(pfdict['x'])

                else:

                    rel_click_x = None

                    self._diagnose("xdisplay_harvest: missing or invalid x-location for mouseclick\n")

                if 'y' in pfdict and re.match('^\d+$', pfdict['y']):

                    rel_click_y = int(pfdict['y'])

                else:

                    rel_click_y = None

                    self._diagnose("xdisplay_harvest: missing or invalid y-location for mouseclick\n")

                if rel_click_x and rel_click_y:

                    d = _xwrap._XOpenDisplay(self.virtual_display)

                    (rc, root, rel_x, rel_y, width, height, border, depth) = _xwrap._XGetGeometry(d, idmatch[0])

                    abs_click_x = rel_x + rel_click_x
                    abs_click_y = rel_y + rel_click_y

                    self._diagnose("xdisplay_harvest: mouseclick at relative click %d,%d; relative window coords "
                                   "%d,%d; absolute click %d,%d\n" %
                                   (rel_click_x, rel_click_y, rel_x, rel_y, abs_click_x, abs_click_y))

                    rc = _xwrap._XTestFakeMotionEvent(d, 0, abs_click_x, abs_click_y, _xwrap.CurrentTime)

                    self._diagnose("xdisplay_harvest: mousemove rc %d\n" % rc)

                    if rc != 0:

                        rc = _xwrap._XTestFakeButtonEvent(d, 1, True, _xwrap.CurrentTime)

                        self._diagnose("xdisplay_harvest: mousedown rc %d\n" % rc)

                        if rc != 0:

                            rc = _xwrap._XTestFakeButtonEvent(d, 1, False, _xwrap.CurrentTime)

                            self._diagnose("xdisplay_harvest: mouseup rc %d\n" % rc)

                            if rc == 0:

                                self._diagnose("xdisplay_harvest: mouseup failed\n")

                        else:

                            self._diagnose("xdisplay_harvest: mouseclick failed\n")

                    else:

                        self._diagnose("xdisplay_harvest: mousemove failed\n")

                    # Move mouse back to neutral so it doesn't invoke balloon help etc.

                    rc = _xwrap._XTestFakeMotionEvent(d, 0, rel_x, rel_y, _xwrap.CurrentTime)

                    if rc == 0:

                        self._diagnose("xdisplay_harvest: mousemove back to neutral position failed\n")

                    _xwrap._XCloseDisplay(d)

            else:

                self._diagnose("xdisplay_harvest: unknown action '%s', ignoring\n" % pfdict['action'])

        self._diagnose("xdisplay_harvest: Launching xwd on idmatch %s\n" % xwinid)

        (tmpfh, tmpname) = tempfile.mkstemp()
        os.close(tmpfh)

        xwd = subprocess.check_output(["xwd", "-nobdrs", "-display", self.virtual_display, "-id", xwinid, "-out", tmpname])

        self._diagnose("xdisplay_harvest: Launched xwd to retrieve Xwindow image into filename " + tmpname + "\n")

        try: 
            im = Image.open(tmpname)
        except OSError as ex:
            raise RuntimeError("RtcacheEngine: Failed to open temporary image file '" + tmpname + "' with Image.open()\n") from ex

        os.unlink(tmpname)

        im = im.convert("RGB")

        tempfl = '/tmp/xwindow_harvest_dump_%s_%d_%d.somefmt' % (self.streamname, os.getuid(), os.getpid())

        im.save(tempfl, format=self.image_format)

        self.inform("xdisplay_harvest: finished image harvest at '%s'\n" % (stock.epoch2str(stock.now(), "%D %T %Z")))

        return tempfl

    def xdisplay_imagefile2orb(self, imagefile, orbfd):

        if not os.path.exists(imagefile):

            raise RuntimeError("Image file '%s' does not exist, can't put on orbserver for stream name '%s'; "
                               "attempting to continue\n" % (imagefile, self.streamname))

        f = open(imagefile, 'rb')
        tblock = f.read()
        f.close()

        output_srcname = self.streamname + "/MIME/image/gif"

        try:
            orbfd.put(output_srcname, stock.now(), tblock)
        except:
            raise RuntimeError("Failed (%s) to put image for stream name '%s' on orbserver; attempting to continue\n" %
                               (sys.exc_info()[0], self.streamname))
        else:
            self.inform("xdisplay_imagefile2orb: put image file '%s' on orb as '%s' at '%s'\n" %
                        (imagefile, output_srcname, stock.epoch2str(stock.now(), "%D %T %Z")))

    def xdisplay_killcmd(self):

        try:

            self.subproc.terminate()

            self.subproc.wait()

            if self.subproc.returncode is None:

                self.inform("Failed to subprocess.terminate() subprocess %d, trying subprocess.kill()" %
                            self.subproc.pid)

                self.subproc.kill()

                self.subproc.poll()

                if self.subproc.returncode is None:

                    self.inform("Failed to subprocess.kill() subprocess %d, trying os.kill(9)" % self.subproc.pid)

                    os.kill(self.subproc.pid, 9)

                    self.subproc.poll()

                    if self.subproc.returncode is None:

                        self.inform("Failed to os.kill(9) subprocess %d, giving up" % self.subproc.pid)

                    else:

                        self._diagnose("xdisplay_killcmd: Killed subprocess command via 'kill -9'\n")

                else:

                    self._diagnose("xdisplay_killcmd: Killed subprocess command via subprocess.kill()\n")

            else:

                self._diagnose("xdisplay_killcmd: Killed subprocess command via subprocess.terminate()\n")

        except:

            self.inform("Failed entirely to kill subprocess %d" % self.subproc.pid)

        rc = self.subproc.wait()

        self._diagnose("xdisplay_killcmd: wait() on subprocess %d returned %d; returncode %d" %
                       (self.subproc.pid, rc, self.subproc.returncode))

    def xdisplay_shutdown(self):

        self.xvfb.poll()

        if self.xvfb.returncode is not None:

            self._diagnose("xdisplay_shutdown: Xvfb is not running")

            return None

        try:
            self.xvfb.kill()
        except:
            self._diagnose("xdisplay_shutdown: Xvfb quit")

        self.xvfb.wait()

        if self.xvfb.returncode is not None:
            self._diagnose("xdisplay_shutdown: Xvfb is no longer running, return code %d" % self.xvfb.returncode)
        else:
            raise RuntimeError("xdisplay_shutdown: Failed to kill Xvfb with pid %d" % self.xvfb.pid)

        lockfile = '/tmp/.X' + self.virtual_display[1:] + '-lock'

        if os.path.exists(lockfile):

            os.unlink(lockfile)

        socketfile = '/tmp/.X11-unix/X' + self.virtual_display[1:]

        if os.path.exists(socketfile):

            os.unlink(socketfile)

        return None

    def xdisplay_is_running(self):

        self._diagnose("xdisplay_is_running: Testing run status of Xvfb pid %d and subproc pid %d" %
                       (self.xvfb.pid, self.subproc.pid))

        self.xvfb.poll()

        self.subproc.poll()

        if self.xvfb.returncode is not None:

            self.inform("xdisplay_is_running: Xvfb is not running (return code '%d') at '%s'\n" %
                        (self.xvfb.returncode, stock.epoch2str(stock.now(), "%D %T %Z")))
            return False

        elif self.subproc.returncode is not None:

            self.inform("xdisplay_is_running: graphical client is not running (return code '%d') at '%s'\n" %
                        (self.subproc.returncode, stock.epoch2str(stock.now(), "%D %T %Z")))
            return False

        else:

            self._diagnose("xdisplay_is_running: Xvfb pid %d and subproc pid %d are running" %
                           (self.xvfb.pid, self.subproc.pid))
            return True

    def _init_engine_xwindow(self):

        expected = ['command',
                    'command_env',
                    'image_format',
                    'startup_sleep_sec',
                    'virtual_display',
                    'virtual_screen_geometry',
                    'window_name']

        for param in expected:

            if param not in self.pfdict_config:

                raise RuntimeError("RtcacheEngine: Failed to start xwindow engine for stream name '%s': "
                                   "missing parameter '%s'" % (self.streamname, param))

        if 'virtual_screen_number' in self.pfdict_config:

            self.virtual_screen_number = self.pfdict_config['virtual_screen_number']

            if self.virtual_screen_number == "":

                self.virtual_screen_number = "0"
        else:
            self.virtual_screen_number = "0"

        if 'xwindow_restart_sec' in self.pfdict_config:

            self.xwindow_restart_sec = int(self.pfdict_config['xwindow_restart_sec'])

        else:

            self.xwindow_restart_sec = 0

        self.orbfd = orb.orbopen(self.outputorb, "w")

    def _startup_xwindow(self):

        self.xdisplay_setup(self.virtual_display,
                            self.virtual_screen_number,
                            self.virtual_screen_geometry,
                            self.startup_sleep_sec)

        self.xdisplay_launchcmd(self.command)

        self.latest_startup_time = stock.now()

    def _update_xwindow(self, force=False):

        if force or (self.xwindow_restart_sec > 0 and
                     ((stock.now() - self.latest_startup_time) > self.xwindow_restart_sec)):
            if force:
                self.inform("_update_xwindow: Restarting X display and command for streamname '%s'" % self.streamname)
            else:
                self.inform("_update_xwindow: Restarting X display and command for streamname '%s', " %
                            self.streamname + "periodic restart since current time '%s' is more than '%s' past last "
                                              "start of '%s'" % (stock.epoch2str(stock.now(), "%D %T %Z"),
                                                                 stock.strtdelta(self.xwindow_restart_sec),
                                                                 stock.epoch2str(self.latest_startup_time, "%D %T %Z")))

            self.xdisplay_killcmd()
            self.xdisplay_shutdown()

            self.xdisplay_setup('auto',
                                self.virtual_screen_number,
                                self.virtual_screen_geometry,
                                self.startup_sleep_sec)

            self.xdisplay_launchcmd(self.command)

            self.inform("_update_xwindow: X display and command restarted for streamname '%s'" % self.streamname)

            self.latest_startup_time = stock.now()

    def _harvest_xwindow(self, pfdict_request):

        tmpfl = self.xdisplay_harvest(pfdict=pfdict_request)

        self.xdisplay_imagefile2orb(tmpfl, self.orbfd)

        os.unlink(tmpfl)

    def _shutdown_xwindow(self):

        self.xdisplay_killcmd()

        self.xdisplay_shutdown()

        self.inform("_shutdown_xwindow: X display decommissioned for streamname '%s'" % self.streamname)

    def __init__(self, streamname, pfdict_config):
        """ Initialize a new RtcacheEngine object
        :param streamname:
        :param pfdict_config:
        """

        if streamname in RtcacheEngine.streams:
            raise RuntimeError("RtcacheEngine: Failed to initialize new object under stream name '%s': "
                               "named stream already exists" % streamname)
        else:
            RtcacheEngine.streams[streamname] = self

        self.streamname = streamname
        self.pfdict_config = pfdict_config

        for param in list(self.pfdict_config.keys()):

            cmd = "self.%s = self.pfdict_config['%s']" % (param, param)

            exec(cmd)

        if 'diagnostics' in pfdict_config and stock.yesno(pfdict_config['diagnostics']):
            self.debug = True

        else:
            self.debug = False

        if ('verbose' in pfdict_config and stock.yesno(pfdict_config['verbose'])) or self.debug is True:
            self.verbose = True

        else:
            self.verbose = False

        if 'outputorb' not in pfdict_config:
            raise RuntimeError("RtcacheEngine: parameter 'outputorb' missing for stream name '%s'" % streamname)

        if 'enginetype' not in pfdict_config:
            raise RuntimeError("RtcacheEngine: parameter 'enginetype' missing for stream name '%s'" % streamname)

        else:

            if pfdict_config['enginetype'] == 'diy':

                self._init_engine_diy()

            elif pfdict_config['enginetype'] == 'xwindow':

                self._init_engine_xwindow()

            else:

                raise RuntimeError("RtcacheEngine: enginetype for stream name '%s' not understood: "
                                   "Invalid enginetype '%s'" % (streamname, pfdict_config['enginetype']))

        self.inform("Initialized '%s' engine" % pfdict_config['enginetype'])

    def __del__(self):

        if hasattr(self, 'streamname') and self.streamname in RtcacheEngine.streams:
            del RtcacheEngine.streams[self.streamname]

    def startup(self):

        self._diagnose("startup: Executing startup function")

        if self.enginetype == 'diy':

            self._startup_diy()

        elif self.enginetype == 'xwindow':

            self._startup_xwindow()

        else:

            raise RuntimeError("RtcacheEngine: unexpected unknown enginetype '%s': Skipping startup function call" %
                               self.enginetype)

    def update(self):

        self._diagnose("update: Executing update function")

        if self.enginetype == 'diy':

            self._update_diy()

        elif self.enginetype == 'xwindow':

            self._update_xwindow()

        else:

            raise RuntimeError("RtcacheEngine: unexpected unknown enginetype '%s': Skipping update function call" %
                               self.enginetype)

    def harvest(self, pfdict_request):

        self._diagnose("harvest: Executing harvest function")

        if self.enginetype == 'diy':

            self._harvest_diy(pfdict_request)

        elif self.enginetype == 'xwindow':

            self._harvest_xwindow(pfdict_request)

        else:

            raise RuntimeError("RtcacheEngine: unexpected unknown enginetype '%s': Skipping harvest function call" %
                               self.enginetype)

    def shutdown(self):

        self._diagnose("shutdown: Executing shutdown function")

        if self.enginetype == 'diy':

            self._shutdown_diy()

        elif self.enginetype == 'xwindow':

            self._shutdown_xwindow()

        else:

            raise RuntimeError("RtcacheEngine: unexpected unknown enginetype '%s': Skipping shutdown function call" %
                               self.enginetype)
