import threading
import os
import re

from twisted.web import http, server
from twisted.web.resource import Resource
from twisted.internet import defer
from twisted.python import log

import antelope.orb as orb
import antelope.stock as stock
import antelope.Pkt as Pkt
import antelope.elog as elog
import antelope.brttpkt as brttpkt

defer.setDebugging(True)

os.environ['ANTELOPE_PYTHON_GILRELEASE'] = '1'


class RtcacheResourceStream(object):

    def __init__(self, streamname, pfdict):

        # Inputs

        self.streamname = streamname
        self.pfdict = pfdict
        self.orbname = self.pfdict['orbname']
        self.orbpost = self.pfdict['orbpost']
        self.maxwait_sec = float(self.pfdict['maxwait_sec'])
        self.expiration_sec = float(self.pfdict['expiration_sec'])

        # Hard-wires

        self.tafter = -1.0         # Negative for most recent read-head position
        self.queuesize = 100       # Maximum number of packets in queue
        self.reject = None

        # Derivatives

        self.select = self.streamname + "/.*"
        self.rtcacherequest_srcname = "%s/pf/rtcacherequest" % self.pfdict['rtcache_targetname']
        self.timeout = self.maxwait_sec

        # State

        self.shutdown = False
        self.reapthread = None
        self.mimetype = None
        self.packet = None
        self.packet_time = -1
        self.rtcacherequest_time = None
        self.streamrequest_time = -1
        self.pending_streamrequests = list()    # pending deferreds returned by this object

        RtcacheResource.diagnose(self, "Initializing RtcacheResourceStream", self.streamname)

        self.orbrcvthr = brttpkt.OrbreapThr(self.orbname,
                                            self.select,
                                            self.reject,
                                            self.tafter,
                                            self.timeout,
                                            self.queuesize)

    def startreap(self):

        self.reapthread = threading.Thread(target=self.reap)

        self.reapthread.daemon = True

        self.reapthread.start()

    def stopreap(self):

        RtcacheResource.diagnose(self, "Stopping reap in RtcacheResourceStream", self.streamname)

        self.shutdown = True

    def stopstream(self):

        RtcacheResource.diagnose(self, "Shutting down RtcacheResourceStream", self.streamname)

        self.reapthread.join()

        RtcacheResource.diagnose(self, "RtcacheResourceStream is shut down", self.streamname)

        self.orbrcvthr.destroy()

    def reap(self):

        RtcacheResource.diagnose(self, " Starting RtcacheResourceStream", self.streamname)

        while not self.shutdown:

            RtcacheResource.diagnose(self, "Starting another reap loop", self.streamname)

            while len(self.pending_streamrequests) > 0 and \
                    self.pending_streamrequests[0].streamrequest_time < stock.now() - self.maxwait_sec:

                d = self.pending_streamrequests.pop(0)

                RtcacheResource.diagnose(self, "Flushing pending request %r" % d, self.streamname)

                d.errback(Exception("Wait time %f sec exceeds maximum allowed of %f sec" %
                                    (stock.now() - d.streamrequest_time, self.maxwait_sec)))

            RtcacheResource.diagnose(self, "Waiting for packet on orbrcvthr %r" % self.orbrcvthr, self.streamname)

            try:

                (pktid, srcname, pkttime, packet) = self.orbrcvthr.get()

                RtcacheResource.diagnose(self, "Got response from orbrcvthr %r" % self.orbrcvthr, self.streamname)

            except (brttpkt.NoData, brttpkt.Timeout):

                RtcacheResource.diagnose(self, "No data from orbrcvthr %r" % self.orbrcvthr, self.streamname)
                continue

            if self.rtcacherequest_time is not None and \
                    (stock.now() - self.rtcacherequest_time) > self.maxwait_sec:
                self.rtcacherequest_time = None

            # If problems occur we may need a write lock here

            if re.match(".*MIME/image/gif", srcname):
                try:
                    self.mimetype = "image/gif"
                    self.packet = packet
                    self.packet_time = float(pkttime)
                    RtcacheResource.diagnose(self, "Retrieved a GIF packet of len %d" % len(self.packet),
                                             self.streamname)
                except:
                    RtcacheResource.diagnose(self, "Failed processing GIF packet", self.streamname)

            elif re.match(".*/pf.*", srcname):

                RtcacheResource.diagnose(self, "Retrieved a /pf packet", self.streamname)

                try:

                    pkt = Pkt.Packet(srcname, pkttime, packet)

                    if pkt.type.content != Pkt.Pkt_pf:

                        RtcacheResource.diagnose(self,
                                                 "Nominal /pf packet ('%s') does not resolve to Pkt_pf content -- "
                                                 "not updating cache" % srcname,
                                                 self.streamname)
                        continue

                    else:
                        self.mimetype = "text/html"
                        self.packet = pkt.pf.__str__()
                        self.packet_time = float(pkttime)
                except:
                    RtcacheResource.diagnose(self, "Failed to unstuff '%s' packet -- not updating cache" % srcname,
                                             self.streamname)
                    continue

            else:

                RtcacheResource.diagnose(self, "Packet type of '%s' not recognized -- not updating cache" % srcname,
                                         self.streamname)
                continue

            if self.rtcacherequest_time is not None:
                self.rtcacherequest_time = None

            if (stock.now() - self.packet_time) < self.expiration_sec:

                RtcacheResource.diagnose(self, "Reviewing pending requests, succeeding with currently valid packet")

                for d in self.pending_streamrequests:

                    d.mimetype = self.mimetype
                    d.packet_time = self.packet_time
                    d.packet = self.packet

                    d.callback("Succeeded")

                self.pending_streamrequests = list()

            else:

                RtcacheResource.diagnose(self, "Packet is already expired (%f seconds old; max is %f)" %
                                         (stock.now() - self.packet_time, self.expiration_sec))

        return

    def get_streamobj(self, webrequest):

        RtcacheResource.diagnose(self, "Received streamobj request", self.streamname)

        if (self.packet is not None) and \
           (self.packet_time > stock.now() - self.expiration_sec) and \
           (self.rtcacherequest_time is None or self.packet_time > self.rtcacherequest_time) and \
           (b'action' not in webrequest.args):

            d = defer.succeed("Succeeded Immediately")

            d.pause()

            d.streamname = self.streamname

            d.streamrequest_time = stock.now()

            d.webrequest = webrequest

            d.mimetype = self.mimetype
            d.packet_time = self.packet_time
            d.packet = self.packet

        else:

            if (not self.rtcacherequest_time) or b'action' in webrequest.args:

                RtcacheResource.inform(self, "Posting rtcacherequest parameter-file packet " +
                                       "to orbserver '%s' under srcname '%s'\n" % (self.orbname,
                                                                                   self.rtcacherequest_srcname),
                                       self.streamname)

                pkt = Pkt.Packet()
                pkt.time = stock.now()
                pkt.srcname = Pkt.SrcName(self.rtcacherequest_srcname)
                pkt.type_suffix = 'pf'

                pf = stock.ParameterFile(stock.PFARR)

                for key in list(webrequest.args.keys()):

                    pf[key.decode('utf-8')] = webrequest.args[key][0].decode('utf-8')

                pkt.pf = pf

                (type, packet, srcname, time) = pkt.stuff()

                try:
                    self.orbpost.put(srcname, time, packet)
                    self.rtcacherequest_time = stock.now()
                except:
                    elog.complain("RtcacheResourceStream: stream '%s': Failed to put request on orb " %
                                  self.streamname)

            RtcacheResource.diagnose(self, "Returning deferred result from get_streamobj", self.streamname)

            d = defer.Deferred()

            d.pause()

            d.streamname = self.streamname

            d.streamrequest_time = stock.now()

            d.webrequest = webrequest

            self.pending_streamrequests.append(d)

        return d


class RtcacheResource(Resource):

    def __init__(self, pfdict):

        Resource.__init__(self)

        self.rrstreams = dict()

        self.pfdict = pfdict

        # orb module (getting used here) generated by scripted/orb define orbopen which returns the object
        # The underlying C call (_orb._orbopen) returns an int
        # type(self.pfdict['orbpost']) is scripted/orb:Orb()
        self.pfdict['orbpost'] = orb.orbopen(self.pfdict['orbname'], "w")

        if self.pfdict['orbpost'] is None:
            elog.die("Failed to open orb '%s' to write requests. Bye\n" % self.pfdict['orbname'])

    def __del__(self):

        if self.pfdict['orbpost'] is not None:
            self.pfdict['orbpost'].close()

        for rrs in self.rrstreams:
            rrs.stopreap()

        for rrs in self.rrstreams:
            rrs.stopstream()

        self.rrstreams = dict()

    @staticmethod
    def inform(obj, message, streamname=None):

        mytype = str(type(obj))
        mytype = re.sub('class', '', mytype)
        mytype = re.match('[a-zA-Z]+', mytype)

        if streamname is None:

            streamname = 'N/A'

        if 'verbose' in obj.pfdict and obj.pfdict['verbose']:

            elog.notify("%s: stream '%s': %s" % (mytype, streamname, message))

    @staticmethod
    def diagnose(obj, message, streamname=None):

        mytype = str(type(obj))
        mytype = re.sub('class', '', mytype)
        mytype = re.match('[a-zA-Z]+', mytype)

        if streamname is None:

            streamname = 'N/A'

        if 'diagnostics' in obj.pfdict and obj.pfdict['diagnostics']:

            elog.notify("%s: stream '%s': Diagnostic: %s " % (mytype, streamname, message))

            elog.flush(True, 0)

    def _noop(self, input=None):

        # No-op in attempt to prevent hanging deferreds

        return

    def _render_success(self, result, d):

        RtcacheResource.diagnose(self,
                                 "_render_success called, result is: %r with %s and image size %d, image age %f sec\n" %
                                 (result, d.mimetype, len(d.packet), stock.now() - d.packet_time),
                                 d.streamname)

        d.webrequest.setHeader('content-type', d.mimetype)

        d.webrequest.write(d.packet)

        d.webrequest.finish()

        return "_render_success succeeded for streamname '%s'" % d.streamname

    def _render_failure(self, failure, deferred=None, webrequest=None, streamname=None):

        if deferred is not None:

            webrequest = deferred.webrequest
            streamname = deferred.streamname

        if streamname is None:

            streamname = 'None'

        RtcacheResource.diagnose(self, "_render_failure called, failure is: %r: %s\n" %
                                 (failure, str(failure)), streamname)

        webrequest.setResponseCode(http.INTERNAL_SERVER_ERROR)

        if not webrequest._disconnected:

            webrequest.finish()

        return "Request terminated"

    def _responseFailed(self, err, webrequest):

        if b'streamname' not in webrequest.args:
            streamname = "Unknown stream"
        else:
            streamname = webrequest.args[b'streamname'][0].decode('utf-8')

        RtcacheResource.diagnose(self, "Connection broken, abandoning webrequest: %r" % err)

        return "Finished responding in _responseFailed"

    def render_GET(self, webrequest):

        webrequest.notifyFinish().addErrback(self._responseFailed, webrequest)

        if b'streamname' not in webrequest.args:

            elog.complain("RtcacheResource: webrequest args %s" % str(webrequest.args) )

            self._render_failure(Exception("No streamname argument"), webrequest=webrequest, streamname='NONE')

            return server.NOT_DONE_YET

        else:

            streamname = webrequest.args[b'streamname'][0].decode('utf-8')

        RtcacheResource.diagnose(self, "received webrequest to render cache resource\n", streamname)

        if streamname not in self.rrstreams:

            self.rrstreams[streamname] = RtcacheResourceStream(streamname, self.pfdict)

            self.rrstreams[streamname].startreap()

        try:

            webrequest.deferred = self.rrstreams[streamname].get_streamobj(webrequest)

            webrequest.deferred.addCallback(self._render_success, webrequest.deferred)
            webrequest.deferred.addErrback(self._render_failure, webrequest.deferred)

            if 'diagnostics' in self.pfdict and self.pfdict['diagnostics']:

                webrequest.deferred.addBoth(log.err)

            else:

                webrequest.deferred.addBoth(self._noop)

            webrequest.deferred.unpause()

        except Exception as e:

            self._render_failure(e, webrequest=webrequest, streamname=streamname)

        return server.NOT_DONE_YET
