
import sys
import os
import re

from socket import gethostbyaddr

from twisted.web import http, proxy, resource, rewrite, script, static

import antelope.datascope as datascope
import antelope.stock as stock
import antelope.elog as elog

import rtwebserver.config as config

import rtcache.rtcacheresource as rtcacheresource

class RtwebserverWrapper(resource.Resource):

    def __init__(self, orig):

        resource.Resource.__init__(self)

        if config.diagnostics: 

            elog.notify( "DEBUG RtWebserverWrapper: initialized wrapper" )

        self.resource = orig

    def getChild( self, path, request ):

        if config.diagnostics: 

            elog.notify( "DEBUG RtWebserverWrapper: getChild called to return '" + str(request) + "'" )

        return self.resource.getChild( path, request )

    def render(self, request):

        if config.diagnostics: 

            elog.notify( "DEBUG RtWebserverWrapper: render called to render '" + str(request) + "'" )

        if config.profile:

            tstart = stock.now()

            elog.notify( "PROFILE RtwebserverWrapper start rendering '%r' at '%s'" % ( request.uri, stock.strtime( tstart ) ) )

        peer = request.transport.getPeer().host

        url = request.uri

        (allow, rstrict) = stock.legit_ip( config.pfname, peer )

        request.rstrict = rstrict

        if not allow:

            elog.complain( "Declining access to '%s' for '%s': IP address is " % ( url, peer ) +
                           "either not approved in 'valid_ip_addresses' table, " +
                           "or rejected in 'reject_ip_addresses' table, of parameter-file\n" ) 

            request.setResponseCode( http.FORBIDDEN )

            if 'phrases' in config.sitedict and 'access_forbidden' in config.sitedict['phrases']:

                retstring = config.sitedict['phrases']['access_forbidden']

            else:

                retstring = "<h1>Access Forbidden</h1>"

        else:
            
            retstring = self.resource.render(request)

        notedresponse = "%d %s" % ( request.code, http.RESPONSES[request.code].decode('utf-8') )

        if config.dbtrack:

            dbreq = config.dbtrack.lookup( table = 'requests' )

            try: 

                rc = dbreq.find( "url == '%s' && peer == '%s'" % ( url, peer ) )

            except datascope.DbfindEnd:

                try: 

                    peername = gethostbyaddr( peer )[0]

                    dbreq.record = rc

                    ( nreqs, ) = dbreq.getv( 'nreqs' )

                    dbreq.putv( ('time', stock.now()),
                             ('nreqs', nreqs + 1), 
                             ('response', notedresponse) )

                except:

                    peername = "-"

                    try:

                        rc = dbreq.addv( ('url', url.decode('utf-8')),
                                         ('peer', peer),
                                         ('peername', peername),
                                         ('time', stock.now()),
                                         ('nreqs', 1),
                                         ('response', notedresponse) )

                    except:

                        elog.complain( "Failed to execute dbreq.addv call" )

            except datascope.DbfindError:

                elog.complain( "Failed to add url '%s', peer '%s' to tracking database" % ( url, peer ) )

            dbpeer = config.dbtrack.lookup( table = 'peers' )

            try: 

                rc = dbpeer.find( "peer == '%s'" % ( peer ) )

                dbpeer.record = rc

                ( nreqs, ) = dbpeer.getv( 'nreqs' )

                dbpeer.putv( ('time', stock.now()),
                         ('nreqs', nreqs + 1) )

            except datascope.DbfindEnd:

                try: 

                    peername = gethostbyaddr( peer )[0]

                except:

                    peername = "-"

                try:
                    rc = dbpeer.addv( ('peer', peer),
                                      ('peername', peername),
                                      ('time', stock.now()),
                                      ('nreqs', 1) )
                except:
                    elog.complain( "Failed to execute dbpeer.addv call" )

            except datascope.DbfindError:

                elog.complain( "Failed to update peer '%s' in tracking database" % peer )

        if config.profile:

            elog.notify( "PROFILE RtwebserverWrapper finish rendering '%r', elapsed %s" % ( request.uri, stock.strtdelta( stock.now() - tstart ) ) )

        if type(retstring) == str:

            retstring = retstring.encode('utf-8')

        return retstring


class RtwebserverResource(resource.Resource):

    @staticmethod
    def _rewrite(request):
        # elog.notify( "SCAFFOLD Rewriting request, current host is '%s'\n" % request.getHost() )
        # request.setHost('www.brtt.com',80)
        # request.setHeader('Content-Location', 'http://www.marble.com:8008/moo')
        # request.setHeader('Content-Base', 'http://www.marble.com:8008/moo')
        # elog.notify( "SCAFFOLD Rewritten host is '%s'\n" % request.getHost() )
        pass

    def _putpage(self, key, assettype, asset):

        if config.diagnostics: 

            elog.notify( "DEBUG RtWebserverResource: _putpage called to handle asset '" + key + "' as '" + str(asset) + "'" )
            print(self.listNames())

        if re.match(".*\s+index\s*$", asset):

            isroot = True

            asset = re.sub("\s+index\s*$", "", asset)

        else:

            isroot = False

        if assettype == "file":

            if re.match(".*\s+.*", asset):

               (asset, mimetype) = asset.split()

            else:

                mimetype = "text/html"

            self.putChild(key.encode('utf-8'), RtwebserverWrapper( static.File(asset, defaultType=mimetype) ) )

            if( isroot ):
                
                self.putChild(b'', RtwebserverWrapper( static.File(asset, defaultType=mimetype) ) )

        elif assettype == "pf":
        
            if re.match(".*\s+.*", asset):

                (asset, mimetype) = asset.split()

            else:

                mimetype = "text/plain"

            if asset in config.sitedict:

                content = config.sitedict[asset]

                self.putChild(key.encode('utf-8'), RtwebserverWrapper( static.Data(content, mimetype) ) )

                if( isroot ):

                    self.putChild(b'', RtwebserverWrapper( static.Data(content, mimetype) ) )

            else:

                elog.complain( "No parameter-file key '%s' in array '%s' of parameter-file '%s'; skipping page '%s'" % (asset, config.sitename, config.pfname, key) )

        elif assettype == "rpy":
            
            s = RtwebserverWrapper( script.ResourceScript(asset, None) )

            self.putChild(key.encode('utf-8'), s)

            if( isroot ):
                
                self.putChild(b'', s)

        elif assettype == "rtcache":

            if asset in config.sitedict:

                pfdict = config.sitedict[asset]

                pfdict['verbose'] = config.verbose

                s = RtwebserverWrapper( rtcacheresource.RtcacheResource( pfdict ) )

                self.putChild(key.encode('utf-8'), s)

                if( isroot ):

                    self.putChild(b'', s)

            else:

                elog.complain( "No parameter-file key '%s' in array '%s' of parameter-file '%s'; skipping page '%s'" % (asset, config.sitename, config.pfname, key) )

        elif assettype == "revproxy":

            if re.match(".*:.*", asset):

                (host, details) = asset.split(':')

                if re.match(".*/.*", details):

                    parts = details.split('/')

                    port = int(parts[0])
                    subpath = parts[1:].join('/')

                else: 

                    port = int(details)
                    subpath = '/'

            else:

                host = asset
                port = 80
                subpath = '/'

            elog.notify( "Setting up reverse proxy of '%s:%s%s' to '/%s'\n" % (host, port, subpath, key) )

            rp = RtwebserverWrapper( proxy.ReverseProxyResource(host, port, subpath) )

            s = rewrite.RewriterResource(rp, RtwebserverResource._rewrite)

            self.putChild(key.encode('utf-8'), s)

            if( isroot ):
                
                self.putChild(b'', s)

        else:

            elog.complain( "Asset type '%s' not understood; ignoring page '%s'" % ( assettype, key ) )

        return

    def __init__(self, tree):

        resource.Resource.__init__(self)

        if 'pages' in tree and 'siteconfig' in tree:

            pages = tree['pages']

            if 'pythonpaths' in tree:

                pythonpaths = tree['pythonpaths']

                for apath in pythonpaths:

                    sys.path.append(apath)

        else:

            pages = tree

        for key in list(pages.keys()):

            page = pages[key]

            if type(page) is dict:

                self.putChild(key.encode('utf-8'), RtwebserverResource(page) )

            elif type(page) is str:

                (assettype, asset) = page.split(':', 1)

                self._putpage(key, assettype, asset)

            else:

                elog.complain( "Page type for page '%s' not understood; ignoring page" % key )

        return

