#   Copyright (c) 2012 Boulder Real Time Technologies, Inc.           
#                                                                     
#   This software module is wholly owned by Boulder Real Time         
#   Technologies, Inc. Any use of this software module without        
#   express written permission from Boulder Real Time Technologies,   
#   Inc. is prohibited.                                               

import os
import shutil
import subprocess

import antelope.orb as orb
import antelope.stock as stock
import antelope.datascope as datascope

class SmCache(object):

    geometry = '1280x1024'
    display = 'auto'
    orbname = ''
    dbname = ''
    filecache = dict()
    orbfd = -1
    cachedir = '/tmp'
    appname_counter = 0

    def __init__( self, engine, display, geometry, outputorb, dbname, cachedir, update_timelimit_sec=0 ):

        self.engine = engine
        self.display = display
        self.geometry = geometry
        self.outputorb = outputorb
        self.dbname = dbname = dbname
        self.cachedir = cachedir
        self.update_timelimit_sec = update_timelimit_sec

        self.engine.xdisplay_setup( self.display, 0, self.geometry, 1 )

        self.orbfd = orb.orbopen( self.outputorb, "w" )

        dbptr = datascope.dbopen( self.dbname, "r" )

        self.dbptr = dbptr.lookup( table = 'pfdisc' )

        self.filecache = dict()

    def _find_alarm_dbrecord( self, pfstringident, pfseq ):

        nrecs = self.dbptr.query( datascope.dbRECORD_COUNT )

        irecord = self.dbptr.find( "pfstringident == '%s' && pfseq == '%s'" % ( pfstringident, pfseq ), 
                                    nrecs, 
                                    True )

        return irecord

    def _create_cachefile( self, cachefile, irecord=-1, pfstringident=None, pfseq=None ):

        if irecord == -1: 

            irecord = self._find_alarm_dbrecord( pfstringident, pfseq )

            if irecord < 0:

                return

        appname = 'smcache_smrspalarm%d' % SmCache.appname_counter

        SmCache.appname_counter += 1

        cmd = "display_smrspalarm -nomenubar -appname %s %s %d" % ( appname, self.dbname, irecord )

        self.engine.xdisplay_launchcmd( cmd )

        try:

            tempfile = self.engine.xdisplay_harvest()

        except RuntimeError:

            self.engine.xdisplay_killcmd()

            raise

        subprocess.call( ["tksend", appname, "exit"], env={'DISPLAY':self.engine.virtual_display,'PATH':os.environ['ANTELOPE']+'/bin','PFPATH':os.environ['ANTELOPE']+'/data/pf:./pf','ANTELOPE':os.environ['ANTELOPE']} )

        key = pfstringident + ":" + str( pfseq )

        shutil.move( tempfile, self.filecache[key] )

    def update( self ):

        tstart = stock.now()
        
        nrecs = self.dbptr.query( datascope.dbRECORD_COUNT )

        for irecord in range(nrecs):

            if self.update_timelimit_sec != 0 and stock.now() - tstart > self.update_timelimit_sec:

                return

            self.dbptr.record = irecord

            (pfstate, pfstringident, pfseq, dfile ) = self.dbptr.getv( 'pfstate', 
                                                                       'pfstringident', 
                                                                       'pfseq', 
                                                                       'dfile' )

            key = pfstringident + ":" + str( pfseq )

            if pfstate[0:5] == "final" and key not in self.filecache:

                self.filecache[key] = self.cachedir + "/" + dfile + ".gif"

                if not os.path.exists( self.filecache[key] ):

                    self._create_cachefile( self.filecache[key], irecord, pfstringident, pfseq )

    def harvest( self, pfstringident, pfseq ):

        key = pfstringident + ":" + pfseq
                
        if key not in self.filecache:

            irecord = self._find_alarm_dbrecord( pfstringident, pfseq )

            if irecord < 0:

                return

            dbptr = self.dbptr

            dbptr.record = irecord

            ( dfile, ) = dbptr.getv( 'dfile' )
        
            self.filecache[key] = self.cachedir + "/" + dfile + ".gif"

        if not os.path.exists( self.filecache[key] ):

            self._create_cachefile( self.filecache[key], -1, pfstringident, pfseq )

        if os.path.exists( self.filecache[key] ):

            self.engine.xdisplay_imagefile2orb( self.filecache[key], self.orbfd )

        else:

            raise RuntimeError( "SmCache.harvest: could not find or create image file '%s' for key '%s'" % ( self.filecache[key], key ) )

    def shutdown( self ): 

        self.engine.xdisplay_shutdown()
