#   Copyright (c) 2012 Boulder Real Time Technologies, Inc.           
#                                                                     
#   This software module is wholly owned by Boulder Real Time         
#   Technologies, Inc. Any use of this software module without        
#   express written permission from Boulder Real Time Technologies,   
#   Inc. is prohibited.                                               

import os
import re
import math
from PIL import Image, ImageDraw


import antelope.orb as orb
import antelope.Pkt as Pkt
import antelope.elog as elog
import antelope.stock as stock
import antelope.coords as coords
import antelope.datascope as datascope

from twisted.web import http

from sm.mapmanager import MapManager
from sm.statusmanager import StatusManager

class SmWeb(object):

    sttsmngr = None
    mpmngr   = None
    stdct    = None
    stafac   = dict()

    pstart   = 0

    statusmanager           = property( lambda s: SmWeb.sttsmngr )
    mapmanager              = property( lambda s: SmWeb.mpmngr )
    station_facilities      = property( lambda s: SmWeb.stafac )
    facilities              = property( lambda s: SmWeb.stdct['siteconfig']['facilities'] )
    colors                  = property( lambda s: SmWeb.stdct['siteconfig']['colors'] )
    centerimage_width       = property( lambda s: SmWeb.stdct['siteconfig']['centerimage_width'] )
    max_normal_latency_sec  = property( lambda s: float( SmWeb.stdct['siteconfig']['max_normal_latency_sec'] ) )
    refresh_sec             = property( lambda s: int( SmWeb.stdct['siteconfig']['refresh_sec'] ) )
    alarmlist_nmax          = property( lambda s: int( SmWeb.stdct['siteconfig']['alarmlist_nmax'] ) )

    def __init__( self, pfdict ):

        if SmWeb.stdct == None:

            SmWeb.stdct = pfdict

        if len( list(SmWeb.stafac.keys()) ) == 0:

            for facility_key in list(self.facilities.keys()):

                if 'lat' not in self.facilities[facility_key]:

                    elog.complain( "Parameter-file does not specify a 'lat' for facility '%s'" % facility_key )
                
                else:

                    self.facilities[facility_key]['lat'] = float( self.facilities[facility_key]['lat'] )

                if 'lon' not in self.facilities[facility_key]:

                    elog.complain( "Parameter-file does not specify a 'lon' for facility '%s'" % facility_key )

                else:

                    self.facilities[facility_key]['lon'] = float( self.facilities[facility_key]['lon'] )

                for sta in self.facilities[facility_key]['stations']:

                    SmWeb.stafac[sta] = facility_key

        if SmWeb.mpmngr == None:

            SmWeb.mpmngr = MapManager( pfdict['siteconfig'] )

        if SmWeb.sttsmngr == None:

            SmWeb.sttsmngr = StatusManager( pfdict['siteconfig'], 
                                                self.station_facilities, 
                                                self.colors,
                                                self.max_normal_latency_sec )

    @classmethod
    def profile_start( cls, request, tag ):

        cls.pstart = stock.now()

        if request != None: 

            uri = request.uri

        else: 

            uri = ''

        print("PROFILE '%s' start rendering '%r' at '%s'" % ( tag, uri, stock.strtime( cls.pstart ) ))

    @classmethod
    def profile_end( cls, request, tag ):

        if request != None: 

            uri = request.uri

        else: 

            uri = ''

        print("PROFILE '%s' finish rendering '%r', elapsed '%s'" % ( tag, uri, stock.strtdelta( stock.now() - cls.pstart ) ))

    def parseargs( self, request, arglist ):

        result = list()

        for arg in arglist: 

            if arg not in request.args:

                elog.complain( "parseargs: couldn't find argument '%s' in request" % arg )

                raise LookupError

            elif arg == 'sta' and request.args[arg][0] not in SmWeb.stafac:

                elog.complain( "parseargs: station '%s' from request is not associated with a facility" % request.args[arg][0] )

                raise LookupError

            else: 

                result.append( request.args[arg][0] )

        return result

    def facility_exists( self, request, facility_key ):

        if facility_key not in self.facilities:

            self.write_notfound( request )

            return False
        
        else:

            return True

    def facility_allowed_access( self, request, facility_key ):

        if request.rstrict == None:
        
            return True

        elif request.rstrict != facility_key:

            elog.complain( "check_facility: Declining access to '%s' because it " % facility_key +
                                "does not match restriction to '%s'" % request.rstrict )

            self.write_forbidden( request )

            return False

        else:

            return True
        
    def facility_ok_or_write_failures( self, request, facility_key ):

        if not self.facility_exists( request, facility_key ):

            return False

        elif not self.facility_allowed_access( request, facility_key ):

            return False

        else:

            return True

    def use_local_timezone( self ):

        return stock.yesno( SmWeb.stdct['siteconfig']['use_local_timezone'] )

    def write_htmlhead( self, request, title, refresh=True ):

        request.write( SmWeb.stdct['phrases']['htmlhead_open'] )

        if refresh and self.refresh_sec != 0:
            request.write( '<meta http-equiv="refresh" content="%d">' % self.refresh_sec )

        request.write( "<title>%s</title>" % title )

    def write_htmlbodystart( self, request, imagemapname=None ):

        request.write( "<body>" )

        if imagemapname:

            imagemap = self.mapmanager.imagemap( imagemapname, self.centerimage_width )

            request.write( imagemap )

        request.write("""<div id="pagecontainer">""")

        request.write( SmWeb.stdct['phrases']['header'] )

        request.write("""
                <div id="content">
                """ )

    def write_htmlbodyfinish( self, request, refresh=True ):

        if refresh and self.refresh_sec != 0:

            if self.use_local_timezone():

                tz = ""
            
            else: 

                tz = None

            timestr = stock.epoch2str( stock.now(), SmWeb.stdct['siteconfig']['time_format'], tz )

            request.write( "<h3><-  Page updated %s (refreshes every %d seconds)  -></h3>\n" % 
                           ( timestr, self.refresh_sec ) )

        request.write("</div> <!-- content -->")

        request.write( SmWeb.stdct['phrases']['footer'] )

        request.write( """
                </div>        <!-- pagecontainer -->
                </body>
                </html>
                """ )

        request.finish()

    def write_mainnav( self, request, activepage=None ):

        request.write( """
                    <div id="nav">
                  <ul id="main_menu">
                """ )

        if activepage == 'overall':
            request.write( """
                  <li><a class="activepage" href="/">Overall</a></li>
                """ )
        else:
            request.write( """
                  <li><a href="/">Overall</a></li>
                """ )

        request.write( """
                  <li class="top_menu"><a href="">Facilities...</a>
                        <ul>
                """ )


        for facility_key in sorted( self.facilities.keys() ):

            request.write( """
                        <li><a href="/facility/%s"><span class="private">%s</span></a></li>
                    """ % ( facility_key, facility_key ) )

        request.write( """
                        </ul>
                </li>
                """ )

        if activepage == 'soh':
            request.write( """
                  <li><a class="activepage" href="/soh">SOH</a></li>
               """ )
        else:
            request.write( """
                  <li><a href="/soh">SOH</a></li>
               """ )

        if activepage == 'rtm':
            request.write( """
                  <li><a class="activepage" href="/rtm">RTM</a></li>
               """ )
        else:
            request.write( """
                  <li><a href="/rtm">RTM</a></li>
               """ )

        request.write( """
                </ul>
                </div>
                """ )

    def write_facnav( self, request, facility_key, activepage=None ):

        request.write( """
                    <div id="facnav">
                  <ul>
                """ )

        if request.rstrict == None:

            request.write( """
                  <li><a href="/">Overall</a></li>
                """ )

        if activepage == 'map':
            request.write( """
                  <li><a class="activepage" href="/facility/%s">Facility Map</a></li>
                """ % facility_key )
        else:       
            request.write( """
                  <li><a href="/facility/%s">Facility Map</a></li>
                """ % facility_key )

        if activepage == 'spectra':
            request.write( """
                  <li><a class="activepage" href="/facspectra/%s">Spectra</a></li>
                """ % facility_key )
        else:       
            request.write( """
                  <li><a href="/facspectra/%s">Spectra</a></li>
                """ % facility_key )

        if activepage == 'soh':
            request.write( """
                  <li><a class="activepage" href="/facsoh/%s">Facility SOH</a></li>
                """ % facility_key )
        else:       
            request.write( """
                  <li><a href="/facsoh/%s">Facility SOH</a></li>
                """ % facility_key )

        request.write( """
                </ul>
                </div>
                """ )

    def write_banner_overall( self, request ):

        ( overallstatus, overallcolor, overallstatusmsg ) = self.statusmanager.status_overall()

        request.write("""
                      <div class="statusbanner">
                      <h2 style="background-color:%s;">Overall Status: %s</h2>
                      </div>
                      """ % 
                      ( overallcolor, overallstatusmsg ) )

    def write_banner_facility( self, request, facility_key ):

        ( facstatus, faccolor, facstatusmsg ) = self.statusmanager.status_facility( facility_key )

        request.write("""
                      <div class="statusbanner">
                      <h2 style="background-color:%s;"><span class="private">%s </span>Facility Status: %s</h2>
                      </div>
                             """ % 
                      ( faccolor, facility_key, facstatusmsg ) )

    def write_banner_station( self, request, sta ):

        ( stastatus, stacolor, stastatusmsg ) = self.statusmanager.status_station( sta )

        request.write("""
                      <div class="statusbanner">
                      <h2 style="background-color:%s;">%s Station Status: %s</h2>
                      </div>
                      """ % 
                      ( stacolor, sta, stastatusmsg ) )

        return stastatus

    def write_alarms( self, request, sta ):

            alarms = SmWeb.sttsmngr.getalarms( sta )
            
            if len( alarms ) <= 0:

                request.write("""
                        <h2>No alarms for station '%s'</h2>
                          """ % sta )

            else:

                if len( alarms ) > self.alarmlist_nmax:

                    alarms = alarms[:self.alarmlist_nmax]

                request.write( """
                        <table class="dbtable">
                        <caption>Alarms for station %s</caption>
                               """ % sta )
                
                request.write( """
                        <tr>
                        <th>Alarm Time</th>
                        <th>Alarm State</th>
                        </tr>
                        """ )

                for alarm in alarms:

                    request.write( "<tr>" )

                    url_pfstringident = re.sub( '/', '%2F', alarm['pfstringident'] )

                    weblink = "/smrspalarm?sta=%s&pfstringident=%s&pfseq=%d" % ( sta, url_pfstringident, alarm['pfseq'] )

                    request.write( """
                            <td style="background-color:%s;"><a href="%s">%s</a></td>
                            <td style="background-color:%s;"><a href="%s">%s</a></td>
                            """ % 
                            ( alarm['color'], weblink, alarm['strtime'], 
                              alarm['color'], weblink, alarm['pfstate'] ) )

                    request.write( "</tr>" )

                request.write( """
                        </table>
                               """ )

    def write_cancelbutton_if_needed( self, request, sta, stastatus, pfstringident, pfseq ):

        ( alarmrow, pfstate ) = SmWeb.sttsmngr.getalarmrow( pfstringident, pfseq )

        if stastatus == 'exceedence' and pfstate == 'final':

            request.write( """
                      <button onClick="window.location=\'/cancelalarm?sta=%s&alarmrow=%d&pfstringident=%s&pfseq=%s\'" id="cancelbutton">Cancel Alarm</button>
                      """ 
                      % ( sta, alarmrow, pfstringident, pfseq ) )

    def write_notfound( self, request ):

        request.setResponseCode( http.NOT_FOUND )

        request.write( "Requested page not found" )

        request.finish()

    def write_forbidden( self, request ):

        request.setResponseCode( http.FORBIDDEN )

        request.write( SmWeb.stdct['phrases']['access_forbidden'] )

        request.finish()

