#   Copyright (c) 2012 Boulder Real Time Technologies, Inc.           
#                                                                     
#   This software module is wholly owned by Boulder Real Time         
#   Technologies, Inc. Any use of this software module without        
#   express written permission from Boulder Real Time Technologies,   
#   Inc. is prohibited.                                               

import os
import re
import math
from PIL import Image, ImageDraw


import antelope.orb as orb
import antelope.Pkt as Pkt
import antelope.elog as elog
import antelope.stock as stock
import antelope.coords as coords
import antelope.datascope as datascope

from twisted.internet import reactor
from twisted.web import http

from . import smweb

class StatusManager(object):

    orbfd                         = -1
    dbptr                         = None
    facilities                    = None
    colors                        = None
    max_normal_latency_sec        = None
    status_refresh_sec            = None
    last_refresh_epoch            = 0
    station_facilities            = None
    station_status                = dict()
    sitedict                      = None

    def __init__(self, pfdict, station_facilities, colors, max_normal_latency_sec):

        StatusManager.sitedict = pfdict

        StatusManager.facilities = pfdict['facilities']

        StatusManager.station_facilities = station_facilities
        StatusManager.colors = colors
        StatusManager.max_normal_latency_sec = max_normal_latency_sec

        if 'status_refresh_sec' not in pfdict:

            elog.die( "StatusManager: Required parameter status_refresh_sec not present" )

        else: 

            StatusManager.status_refresh_sec = int( pfdict['status_refresh_sec'] )

        StatusManager.dbptr = datascope.dbopen( pfdict['alarm_database'], 'r+' )

        StatusManager.dbptr = StatusManager.dbptr.lookup( table='pfdisc' )

        StatusManager.orbfd = orb.orbopen( pfdict['alarm_orb'], 'r&' )

        StatusManager.orbfd.select( pfdict['status_srcname'] )

        self._update()

        return

    def _clear_dataproblems( self ):

        timenow = stock.now()

        try:
        
            ( pktid, srcname, time, packet ) = StatusManager.orbfd.get( orb.ORBNEWEST )
        
        except:

            elog.complain( "StatusManager: Data Problem: Failed to get latest alarm status packet" )

            return

        if timenow - time > StatusManager.max_normal_latency_sec:

            elog.complain( "StatusManager: Data Problem: latest alarm status packet is too latent " +
                                "(%f seconds > %f max_normal_latency_sec)" %
                                ( timenow - time, StatusManager.max_normal_latency_sec ) )

            return

        pkt = Pkt.Packet( srcname, time, packet )

        if pkt.type.content != Pkt.Pkt_pf:

            elog.complain( "StatusManager: Data Problem: latest alarm status packet is not in parameter-file format" )

            return

        pfdict = pkt.pf.pf2dict()

        if 'staprocs' not in pfdict:

            elog.complain( "StatusManager: Data Problem: latest alarm status packet is missing the 'staprocs' array" )

            return

        for key in pfdict['staprocs']:

            if not re.match( "^[a-zA-Z0-9]+_[a-zA-Z0-9]+", key ):

                elog.complain( "StatusManager: Data Problem: orb packet '%s' " % srcname +
                               "contains info for '%s' " % key +
                               "which does not match expected naming conventions ('^[a-zA-Z0-9]+_[a-zA-Z0-9]+'); ignoring '%s'" % key )

                continue

            sta_short = re.match( "^[a-zA-Z0-9]+_([a-zA-Z0-9]+)", key ).group(1)
            sta_long = re.match( "^([a-zA-Z0-9]+_[a-zA-Z0-9]+)", key ).group(1)

            if( sta_short in StatusManager.station_status ):

                sta = sta_short

            elif( sta_long in StatusManager.station_status ):

                sta = sta_long

            else:

                elog.complain( "StatusManager: Data Problem: orb packet '%s' " % srcname +
                               "contains info for '%s' " % key +
                               "but we're not monitoring station '%s' or '%s'; ignoring '%s'" % ( sta_short, sta_long, key ) )

                continue

            if 'spec_data_time' not in pfdict['staprocs'][key]:
             
                elog.complain( "StatusManager: Data Problem: latest alarm status packet " +
                                    "is missing the 'spec_data_time' parameter for " +
                                    "station '%s'" % sta )

                continue

            if pfdict['staprocs'][key]['spec_data_time'] == '-':

                elog.complain( "StatusManager: Data Problem: null spec_data_time in heartbeat packet for station '%s'" % sta )

            elif timenow - float( pfdict['staprocs'][key]['spec_data_time'] ) < StatusManager.max_normal_latency_sec:
            
                if sta in StatusManager.station_status:

                    StatusManager.station_status[sta] = 'normal'
                
                else: 

                    elog.complain( "StatusManager: Data Problem: station '%s' evaluates to 'normal' " % sta +
                                   "but does not exist in station-status array '%r'" % StatusManager.station_status )
                    
            else: 

                elog.complain( "StatusManager: Data Problem: latest alarm status packet " +
                                    "shows 'spec_data_time' parameter for " +
                                    "station '%s' to be too latent " % sta +
                                    "(%f seconds > %s max_normal_latency_sec)" %
                                    ( timenow - float( pfdict['staprocs'][key]['spec_data_time'] ), 
                                      StatusManager.max_normal_latency_sec ) )

    def _update(self, force_unscheduled_run=False):

        rerun_interval_sec = int( StatusManager.status_refresh_sec / 3 )

        if not force_unscheduled_run:

            if ( len( StatusManager.station_status ) > 0 ) and ( ( stock.now() - StatusManager.last_refresh_epoch ) < StatusManager.status_refresh_sec ):

                reactor.callLater( rerun_interval_sec, self._update )

                return

        StatusManager.station_status = dict()

        for sta in list(StatusManager.station_facilities.keys()):

            StatusManager.station_status[sta] = 'dataproblem'

        self._clear_dataproblems()

        db = StatusManager.dbptr

        nrecs = db.query( datascope.dbRECORD_COUNT )

        for i in range( nrecs ):

            db.record = i

            ( srcname, pfstate ) = db.getv( 'srcname', 'pfstate' )

            sta = re.match( "^[A-Z0-9]+_([A-Z0-9]+)", srcname ).group(1)

            if pfstate == 'final':
             
                StatusManager.station_status[sta] = 'exceedence'

        StatusManager.last_refresh_epoch = stock.now()

        if not force_unscheduled_run: 

            reactor.callLater( rerun_interval_sec, self._update )

    def getalarmrow( self, pfstringident, pfseq ):

        db = StatusManager.dbptr.lookup( table='pfdisc' )

        nrecs = db.query( datascope.dbRECORD_COUNT )

        row = db.find( "pfstringident == '%s' && pfseq == %s" % ( pfstringident, pfseq ),
                        nrecs,
                        True )

        if row > 0:

            db.record = row

            ( pfstate, ) = db.getv( 'pfstate' )

        else: 

            pfstate = None

        return ( row, pfstate )

    def getalarms( self, sta ):

        alarms = list()

        StatusManager.dbptr.close()

        StatusManager.dbptr = datascope.dbopen( StatusManager.sitedict['alarm_database'], 'r+' )

        StatusManager.dbptr = StatusManager.dbptr.lookup( table='pfdisc' )

        db = StatusManager.dbptr
        
        db = db.sort( ['time'], reverse = True )

        db = db.subset( "srcname =~ /.*_%s_.*/ && pfstate =~ /final.*/ && isstash == 'n'" % sta )

        nrecs = db.query( datascope.dbRECORD_COUNT )

        for i in range(nrecs):

            alarm = dict()

            db.record = i

            ( pkttime, pfstate, pfstringident, pfseq ) = db.getv( 'time', 'pfstate', 'pfstringident', 'pfseq' )

            alarm['pkttime'] = pkttime
            alarm['pfstate'] = pfstate
            alarm['pfstringident'] = pfstringident
            alarm['pfseq'] = pfseq

            alarmtime = re.match( ".*/([0-9.]+)", pfstringident ).group(1)
            alarm['time'] = float( alarmtime )

            if stock.yesno( smweb.SmWeb.stdct['siteconfig']['use_local_timezone'] ):

                tz = ""

            else:

                tz = None

            alarm['strtime'] = stock.epoch2str( alarm['time'], smweb.SmWeb.stdct['siteconfig']['time_format'], tz )

            if pfstate == 'final-ack':

                alarm['color'] = StatusManager.colors['normal']

            else:

                alarm['color'] = StatusManager.colors['exceedence']

            alarms.append( alarm )

        return alarms

    def status_station(self, sta):

        messages = { 'normal': 'Normal',
                     'exceedence': 'Alarm',
                     'dataproblem': 'Data Problem' }

        if sta not in StatusManager.station_status:

            elog.complain( "StatusManager: Cannot find status for station '%s' (station not listed for any facility)\n" % sta )

            return( "dataproblem", StatusManager.colors['dataproblem'], messages['dataproblem'] )

        statusmsg = messages[StatusManager.station_status[sta]]

        return ( StatusManager.station_status[sta], StatusManager.colors[StatusManager.station_status[sta]], statusmsg )

    def status_facility(self, facility_key):

        exceedences = list()
        dataproblems = list()

        status = "normal"

        for sta in StatusManager.facilities[facility_key]['stations']:

            ( stastatus, stacolor, stastatusmsg ) = self.status_station( sta )

            if stastatus == "exceedence":

                exceedences.append( sta )

                status = "exceedence"

            elif stastatus == "dataproblem":

                dataproblems.append( sta )

                if status != 'exceedence':

                    status = "dataproblem"

        if len( exceedences ) <= 0 and len( dataproblems ) <= 0:

            statusmsg = "Normal"

        else: 

            statusmsg = ""

            if len( exceedences ) > 0:

                if len( exceedences ) == 1:

                    plural = ""

                else:

                    plural = "s"

                statusmsg += "Alarm%s<span class='private'> at %s" % ( plural, ", ".join( exceedences ) ) + "</span>"

            if len( dataproblems ) > 0:

                if len( dataproblems ) == 1:

                    plural = ""

                else:

                    plural = "s"

                if statusmsg != "": 

                    statusmsg += "; "

                statusmsg += "Data Problem%s<span class='private'> at %s" % ( plural, ", ".join( dataproblems ) ) + "</span>"

        color = StatusManager.colors[status]

        return ( status, color, statusmsg )

    def status_overall(self):

        exceedences = list()
        dataproblems = list()

        status = "normal"

        for key in list(StatusManager.facilities.keys()):

            ( facstatus, faccolor, facstatusmsg ) = self.status_facility( key )

            if facstatus == "exceedence":

                exceedences.append( key )

                status = "exceedence"

            elif facstatus == "dataproblem":

                dataproblems.append( key )

                if status != 'exceedence':

                    status = "dataproblem"

        if len( exceedences ) <= 0 and len( dataproblems ) <= 0:

            statusmsg = "Normal"

        else: 

            statusmsg = ""

            if len( exceedences ) > 0:

                if len( exceedences ) == 1:

                    plural = ""

                else:

                    plural = "s"

                statusmsg += "Alarm%s<span class='private'> at " % plural + ", ".join(exceedences) + "</span>"

            if len( dataproblems ) > 0:

                if len( dataproblems ) == 1:

                    plural = ""

                else:

                    plural = "s"

                if statusmsg != "": 

                    statusmsg += "; "

                statusmsg += "Data Problem%s at <span class='private'>" % plural + ", ".join(dataproblems) + "</span>"

        color = StatusManager.colors[status]

        return ( status, color, statusmsg )

    def cancelalarm( self, alarmrow ):

        db = StatusManager.dbptr

        db = db.lookup( table='pfdisc' )

        nrecs = db.query( datascope.dbRECORD_COUNT )

        if type( alarmrow ) != int or alarmrow < 0 or alarmrow >= nrecs:

            elog.complain( "cancelalarm: Invalid alarm row %r" % alarmrow )

            return -1

        db.record = alarmrow

        ( pfstate, ) = db.getv( 'pfstate' )

        if pfstate != 'final':

            elog.complain( "cancelalarm: alarm row %d not cancellable, pfstate is not 'final'" % alarmrow )

            return -1

        db.putv( ('pfstate', 'final-ack') )

        self._update( force_unscheduled_run=True )

        return 0
